import subprocess

import anyscale


# Make sure anyscale is using the vendored Ray
def test_use_vendored_ray_in_cli() -> None:
    # Make sure to use vendored ray when anyscale is used in the CLI.
    assert b"anyscale_ray" in subprocess.check_output(  # noqa: B1
        ["python", "-c", "import anyscale; import ray; print(ray.__path__)"],
        env=anyscale.ANYSCALE_ENV,
    )
    # Make sure to not use vendored ray when anyscale is used as a
    # package (e.g. if we use anyscale.connect).
    assert b"anyscale_ray" not in subprocess.check_output(  # noqa: B1
        ["python", "-c", "import anyscale; import ray; print(ray.__path__)"],
    )



import json
import os
import tempfile

from click import ClickException
import pytest

from anyscale import credentials


def test_validate_credential() -> None:
    with pytest.raises(ClickException):
        credentials._validate_credential("abc")

    with pytest.raises(ClickException):
        credentials._validate_credential("")

    # Should not raise an exception.
    credentials._validate_credential("sss_abcdefghjijklmnop")


def test_credentials_chose_environ_over_folder() -> None:
    old_environ = os.environ.copy()
    try:
        os.environ["ANYSCALE_CLI_TOKEN"] = "sss_os_environ"
        with tempfile.NamedTemporaryFile("w") as temp_credentials_file:
            temp_credentials_file.write(
                json.dumps({"cli_token": "sss_file_credential"})
            )
            temp_credentials_file.flush()
            credentials.CREDENTIALS_FILE = temp_credentials_file.name
            assert credentials.load_credentials() == "sss_os_environ"
    finally:
        os.environ = old_environ  # type: ignore



"""
Commands to interact with the Anyscale Sessions API
"""

from typing import Optional

import click

from anyscale.api import get_anyscale_api_client
from anyscale.formatters import common_formatter
from anyscale.sdk.anyscale_client import (
    CreateSession,
    StartSessionOptions,
    StopSessionOptions,
    TerminateSessionOptions,
    UpdateSession,
)


@click.group(
    "sessions", help="Commands to interact with the Sessions API.",
)
def sessions() -> None:
    pass


@sessions.command(
    name="list", short_help="Lists all Sessions belonging to the Project."
)
@click.argument("project_id", required=True)
@click.option(
    "--count", type=int, default=10, help="Number of projects to show. Defaults to 10."
)
@click.option(
    "--paging-token",
    required=False,
    help="Paging token used to fetch subsequent pages of projects.",
)
def list_sessions(project_id: str, count: int, paging_token: Optional[str],) -> None:
    """Lists all the non-deleted sessions under PROJECT_ID. """

    api_client = get_anyscale_api_client()
    response = api_client.list_sessions(
        project_id, count=count, paging_token=paging_token
    )

    print(common_formatter.prettify_json(response.to_dict()))


@sessions.command(name="get", short_help="Retrieves a Session.")
@click.argument("session_id", required=True)
def get_session(session_id: str) -> None:
    """Get details about the Session with id SESSION_ID"""

    api_client = get_anyscale_api_client()
    response = api_client.get_session(session_id)

    print(common_formatter.prettify_json(response.to_dict()))


@sessions.command(name="create", short_help="Creates a Session.")
@click.argument("name", required=True)
@click.argument("project_id", required=True)
@click.argument("cloud_id", required=True)
@click.argument("cluster_config", required=True)
def create_session(
    name: str, project_id: str, cloud_id: str, cluster_config: str,
) -> None:
    """Creates a Session with NAME, PROJECT_ID, CLOUD_ID, CLUSTER_CONFIG."""

    api_client = get_anyscale_api_client()
    create_data = CreateSession(
        name=name,
        cluster_config=cluster_config,
        project_id=project_id,
        cloud_id=cloud_id,
    )
    response = api_client.create_session(create_data)

    print(common_formatter.prettify_json(response.to_dict()))


@sessions.command(name="update", short_help="Updates a Session.")
@click.argument("session_id", required=True)
@click.argument("cluster_config", required=False)
@click.option(
    "--idle-timeout", required=False, help="Idle timeout (in minutes)", type=int,
)
def update_session(
    session_id: str, cluster_config: Optional[str], idle_timeout: Optional[int]
) -> None:
    """
    Updates Session SESSION_ID with CLUSTER_CONFIG and/or idle-timeout.
    """

    api_client = get_anyscale_api_client()
    update_data = UpdateSession(
        cluster_config=cluster_config, idle_timeout=idle_timeout
    )
    response = api_client.update_session(session_id, update_data)
    print(response)

    print(common_formatter.prettify_json(response.to_dict()))


@sessions.command(name="delete", short_help="Deletes a Session.")
@click.argument("session_id", required=True)
def delete_session(session_id: str) -> None:
    """Delete the Session with id SESSION_ID"""

    api_client = get_anyscale_api_client()
    api_client.delete_session(session_id)


@sessions.command(name="start", short_help="Sets session goal state to Running.")
@click.argument("session_id", required=True)
@click.argument("cluster_config", required=False)
def start_session(session_id: str, cluster_config: Optional[str]) -> None:
    """Sets session goal state to Running
    using the given cluster_config (if specified).
    A session with goal state running will eventually
    transition from its current state to Running, or
    remain Running if its current state is already Running.
    Retrieves the corresponding session operation.
    The session will update if necessary to the given
    cluster_config.

    cluster_config: New cluster config to apply to the Session.
    """

    api_client = get_anyscale_api_client()
    start_session_options = StartSessionOptions(cluster_config=cluster_config)
    response = api_client.start_session(session_id, start_session_options)
    print(common_formatter.prettify_json(response.to_dict()))


@sessions.command(name="stop", short_help="Sets session goal state to Stopped.")
@click.argument("session_id", required=True)
@click.argument("workers_only", required=False, default=False)
@click.argument("keep_min_workers", required=False, default=False)
def stop_session(session_id: str, workers_only: bool, keep_min_workers: bool) -> None:
    """Sets session goal state to Stopped.
    A session with goal state Stopped will eventually
    transition from its current state to Stopped, or
    remain Stopped if its current state is already Stopped.
    Retrieves the corresponding session operation.
    Warning: A session with goal state Terminated may not
    have its goal state changed to Stopped.

    workers_only: Default False. Only destroy the workers when stopping.

    keep_min_workers: Default False. Retain the minimal amount of workers
    specified in the config when stopping.

    """

    api_client = get_anyscale_api_client()
    stop_session_options = StopSessionOptions(
        workers_only=workers_only, keep_min_workers=keep_min_workers
    )
    response = api_client.stop_session(session_id, stop_session_options)
    print(common_formatter.prettify_json(response.to_dict()))


@sessions.command(name="terminate", short_help="Sets session goal state to Terminated.")
@click.argument("session_id", required=True)
@click.argument("workers_only", required=False, default=False)
@click.argument("keep_min_workers", required=False, default=False)
@click.argument("delete", required=False, default=False)
@click.argument("take_snapshot", required=False, default=True)
def terminate_session(
    session_id: str,
    workers_only: bool,
    keep_min_workers: bool,
    delete: bool,
    take_snapshot: bool,
) -> None:
    """Sets session goal state to Terminated.
    A session with goal state Terminated will eventually
    transition from its current state to Terminated, or
    remain Terminated if its current state is already Terminated.
    Retrieves the corresponding session operation.

    workers_only: Default False. Only destroy the workers when stopping.

    keep_min_workers: Default False. Retain the minimal amount of workers
    specified in the config when stopping.

    delete: Default False. Delete the session after terminating.

    take_snapshot: Default True. Takes a snapshot to preserve the state of
    the session before terminating. The state will be restored
    the next time the session is started.
    """

    api_client = get_anyscale_api_client()
    terminate_session_options = TerminateSessionOptions(
        workers_only=workers_only,
        keep_min_workers=keep_min_workers,
        delete=delete,
        take_snapshot=take_snapshot,
    )
    response = api_client.terminate_session(session_id, terminate_session_options)
    print(common_formatter.prettify_json(response.to_dict()))



from typing import Optional


AWS_PROFILE = None


# Global variable that contains the server session token.
CLI_TOKEN: Optional[str] = None

TEST_V2 = False


ANYSCALE_IAM_ROLE_NAME = "anyscale-iam-role"



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class ComputeTemplateConfig(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'cloud_id': 'str',
        'max_workers': 'int',
        'region': 'str',
        'allowed_azs': 'list[str]',
        'head_node_type': 'ComputeNodeType',
        'worker_node_types': 'list[WorkerNodeType]',
        'aws': 'AWSNodeOptions',
        'gcp': 'object',
        'azure': 'object'
    }

    attribute_map = {
        'cloud_id': 'cloud_id',
        'max_workers': 'max_workers',
        'region': 'region',
        'allowed_azs': 'allowed_azs',
        'head_node_type': 'head_node_type',
        'worker_node_types': 'worker_node_types',
        'aws': 'aws',
        'gcp': 'gcp',
        'azure': 'azure'
    }

    def __init__(self, cloud_id=None, max_workers=None, region=None, allowed_azs=None, head_node_type=None, worker_node_types=None, aws=None, gcp=None, azure=None, local_vars_configuration=None):  # noqa: E501
        """ComputeTemplateConfig - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._cloud_id = None
        self._max_workers = None
        self._region = None
        self._allowed_azs = None
        self._head_node_type = None
        self._worker_node_types = None
        self._aws = None
        self._gcp = None
        self._azure = None
        self.discriminator = None

        self.cloud_id = cloud_id
        if max_workers is not None:
            self.max_workers = max_workers
        self.region = region
        if allowed_azs is not None:
            self.allowed_azs = allowed_azs
        self.head_node_type = head_node_type
        self.worker_node_types = worker_node_types
        if aws is not None:
            self.aws = aws
        if gcp is not None:
            self.gcp = gcp
        if azure is not None:
            self.azure = azure

    @property
    def cloud_id(self):
        """Gets the cloud_id of this ComputeTemplateConfig.  # noqa: E501

        The ID of the Anyscale cloud to use for launching sessions.  # noqa: E501

        :return: The cloud_id of this ComputeTemplateConfig.  # noqa: E501
        :rtype: str
        """
        return self._cloud_id

    @cloud_id.setter
    def cloud_id(self, cloud_id):
        """Sets the cloud_id of this ComputeTemplateConfig.

        The ID of the Anyscale cloud to use for launching sessions.  # noqa: E501

        :param cloud_id: The cloud_id of this ComputeTemplateConfig.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and cloud_id is None:  # noqa: E501
            raise ValueError("Invalid value for `cloud_id`, must not be `None`")  # noqa: E501

        self._cloud_id = cloud_id

    @property
    def max_workers(self):
        """Gets the max_workers of this ComputeTemplateConfig.  # noqa: E501

        Desired limit on total running workers for this session.  # noqa: E501

        :return: The max_workers of this ComputeTemplateConfig.  # noqa: E501
        :rtype: int
        """
        return self._max_workers

    @max_workers.setter
    def max_workers(self, max_workers):
        """Sets the max_workers of this ComputeTemplateConfig.

        Desired limit on total running workers for this session.  # noqa: E501

        :param max_workers: The max_workers of this ComputeTemplateConfig.  # noqa: E501
        :type: int
        """

        self._max_workers = max_workers

    @property
    def region(self):
        """Gets the region of this ComputeTemplateConfig.  # noqa: E501

        The region to launch sessions in, e.g. \"us-west-2\".  # noqa: E501

        :return: The region of this ComputeTemplateConfig.  # noqa: E501
        :rtype: str
        """
        return self._region

    @region.setter
    def region(self, region):
        """Sets the region of this ComputeTemplateConfig.

        The region to launch sessions in, e.g. \"us-west-2\".  # noqa: E501

        :param region: The region of this ComputeTemplateConfig.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and region is None:  # noqa: E501
            raise ValueError("Invalid value for `region`, must not be `None`")  # noqa: E501

        self._region = region

    @property
    def allowed_azs(self):
        """Gets the allowed_azs of this ComputeTemplateConfig.  # noqa: E501

        The availability zones that sessions are allowed to be launched in, e.g. \"us-west-2a\". If not specified, any AZ may be used.  # noqa: E501

        :return: The allowed_azs of this ComputeTemplateConfig.  # noqa: E501
        :rtype: list[str]
        """
        return self._allowed_azs

    @allowed_azs.setter
    def allowed_azs(self, allowed_azs):
        """Sets the allowed_azs of this ComputeTemplateConfig.

        The availability zones that sessions are allowed to be launched in, e.g. \"us-west-2a\". If not specified, any AZ may be used.  # noqa: E501

        :param allowed_azs: The allowed_azs of this ComputeTemplateConfig.  # noqa: E501
        :type: list[str]
        """

        self._allowed_azs = allowed_azs

    @property
    def head_node_type(self):
        """Gets the head_node_type of this ComputeTemplateConfig.  # noqa: E501

        Node configuration to use for the head node.   # noqa: E501

        :return: The head_node_type of this ComputeTemplateConfig.  # noqa: E501
        :rtype: ComputeNodeType
        """
        return self._head_node_type

    @head_node_type.setter
    def head_node_type(self, head_node_type):
        """Sets the head_node_type of this ComputeTemplateConfig.

        Node configuration to use for the head node.   # noqa: E501

        :param head_node_type: The head_node_type of this ComputeTemplateConfig.  # noqa: E501
        :type: ComputeNodeType
        """
        if self.local_vars_configuration.client_side_validation and head_node_type is None:  # noqa: E501
            raise ValueError("Invalid value for `head_node_type`, must not be `None`")  # noqa: E501

        self._head_node_type = head_node_type

    @property
    def worker_node_types(self):
        """Gets the worker_node_types of this ComputeTemplateConfig.  # noqa: E501

        A list of node types to use for worker nodes.   # noqa: E501

        :return: The worker_node_types of this ComputeTemplateConfig.  # noqa: E501
        :rtype: list[WorkerNodeType]
        """
        return self._worker_node_types

    @worker_node_types.setter
    def worker_node_types(self, worker_node_types):
        """Sets the worker_node_types of this ComputeTemplateConfig.

        A list of node types to use for worker nodes.   # noqa: E501

        :param worker_node_types: The worker_node_types of this ComputeTemplateConfig.  # noqa: E501
        :type: list[WorkerNodeType]
        """
        if self.local_vars_configuration.client_side_validation and worker_node_types is None:  # noqa: E501
            raise ValueError("Invalid value for `worker_node_types`, must not be `None`")  # noqa: E501

        self._worker_node_types = worker_node_types

    @property
    def aws(self):
        """Gets the aws of this ComputeTemplateConfig.  # noqa: E501

        Fields specific to AWS node types.  # noqa: E501

        :return: The aws of this ComputeTemplateConfig.  # noqa: E501
        :rtype: AWSNodeOptions
        """
        return self._aws

    @aws.setter
    def aws(self, aws):
        """Sets the aws of this ComputeTemplateConfig.

        Fields specific to AWS node types.  # noqa: E501

        :param aws: The aws of this ComputeTemplateConfig.  # noqa: E501
        :type: AWSNodeOptions
        """

        self._aws = aws

    @property
    def gcp(self):
        """Gets the gcp of this ComputeTemplateConfig.  # noqa: E501

        Fields specific to GCP node types.  # noqa: E501

        :return: The gcp of this ComputeTemplateConfig.  # noqa: E501
        :rtype: object
        """
        return self._gcp

    @gcp.setter
    def gcp(self, gcp):
        """Sets the gcp of this ComputeTemplateConfig.

        Fields specific to GCP node types.  # noqa: E501

        :param gcp: The gcp of this ComputeTemplateConfig.  # noqa: E501
        :type: object
        """

        self._gcp = gcp

    @property
    def azure(self):
        """Gets the azure of this ComputeTemplateConfig.  # noqa: E501

        Fields specific to Azure node types.  # noqa: E501

        :return: The azure of this ComputeTemplateConfig.  # noqa: E501
        :rtype: object
        """
        return self._azure

    @azure.setter
    def azure(self, azure):
        """Sets the azure of this ComputeTemplateConfig.

        Fields specific to Azure node types.  # noqa: E501

        :param azure: The azure of this ComputeTemplateConfig.  # noqa: E501
        :type: object
        """

        self._azure = azure

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ComputeTemplateConfig):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ComputeTemplateConfig):
            return True

        return self.to_dict() != other.to_dict()



from contextlib import contextmanager
import datetime
import json
import logging
import os
import subprocess
import sys
import threading
import time
from typing import Any, cast, Dict, Iterator, List, Optional, Tuple, Type
from urllib.parse import quote as quote_sanitize, urlencode, urljoin
import webbrowser

import click
import requests
from requests import Response
import yaml

from anyscale.api import get_api_client
from anyscale.client.openapi_client.api.default_api import DefaultApi
from anyscale.client.openapi_client.models.user_info import UserInfo
from anyscale.client.openapi_client.rest import ApiException
import anyscale.conf
from anyscale.credentials import load_credentials
import anyscale.shared_anyscale_utils.conf as shared_anyscale_conf
from anyscale.shared_anyscale_utils.util import get_container_name


logger = logging.getLogger(__file__)

BOTO_MAX_RETRIES = 5


def confirm(msg: str, yes: bool) -> Any:
    return None if yes else click.confirm(msg, abort=True)


def get_endpoint(endpoint: str, host: Optional[str] = None) -> str:
    return str(urljoin(host or shared_anyscale_conf.ANYSCALE_HOST, endpoint))


def send_json_request_raw(
    endpoint: str,
    json_args: Dict[str, Any],
    method: str = "GET",
    cli_token: Optional[str] = None,
    host: Optional[str] = None,
) -> Response:
    if anyscale.conf.CLI_TOKEN is None and not cli_token:
        anyscale.conf.CLI_TOKEN = load_credentials()

    url = get_endpoint(endpoint, host=host)
    cookies = {"cli_token": cli_token or anyscale.conf.CLI_TOKEN}
    try:
        if method == "GET":
            resp = requests.get(url, params=json_args, cookies=cookies)
        elif method == "POST":
            resp = requests.post(url, json=json_args, cookies=cookies)
        elif method == "DELETE":
            resp = requests.delete(url, json=json_args, cookies=cookies)
        elif method == "PATCH":
            resp = requests.patch(url, data=json_args, cookies=cookies)
        elif method == "PUT":
            resp = requests.put(url, json=json_args, cookies=cookies)
        else:
            assert False, "unknown method {}".format(method)
    except requests.exceptions.ConnectionError:
        raise click.ClickException(
            "Failed to connect to anyscale server at {}".format(url)
        )

    return resp


def send_json_request(
    endpoint: str,
    json_args: Dict[str, Any],
    method: str = "GET",
    cli_token: Optional[str] = None,
    host: Optional[str] = None,
) -> Dict[str, Any]:
    resp = send_json_request_raw(
        endpoint, json_args, method=method, cli_token=cli_token, host=host,
    )

    if not resp.ok:
        if resp.status_code == 500:
            raise click.ClickException(
                "There was an internal error in this command. "
                "Please report this to the Anyscale team at beta@anyscale.com "
                "with the token '{}'.".format(resp.headers["x-trace-id"])
            )

        raise click.ClickException("{}: {}.".format(resp.status_code, resp.text))

    if resp.status_code == 204:
        return {}

    json_resp: Dict[str, Any] = resp.json()
    if "error" in json_resp:
        raise click.ClickException("{}".format(json_resp["error"]))

    return json_resp


def deserialize_datetime(s: str) -> datetime.datetime:
    if sys.version_info < (3, 7) and ":" == s[-3:-2]:
        s = s[:-3] + s[-2:]

    return datetime.datetime.strptime(s, "%Y-%m-%dT%H:%M:%S.%f%z")


def humanize_timestamp(timestamp: datetime.datetime) -> str:
    delta = datetime.datetime.now(datetime.timezone.utc) - timestamp
    offset = float(delta.seconds + (delta.days * 60 * 60 * 24))
    delta_s = int(offset % 60)
    offset /= 60
    delta_m = int(offset % 60)
    offset /= 60
    delta_h = int(offset % 24)
    offset /= 24
    delta_d = int(offset)

    if delta_d >= 1:
        return "{} day{} ago".format(delta_d, "s" if delta_d > 1 else "")
    if delta_h > 0:
        return "{} hour{} ago".format(delta_h, "s" if delta_h > 1 else "")
    if delta_m > 0:
        return "{} minute{} ago".format(delta_m, "s" if delta_m > 1 else "")
    else:
        return "{} second{} ago".format(delta_s, "s" if delta_s > 1 else "")


def get_cluster_config(config_path: str) -> Any:
    with open(config_path) as f:
        cluster_config = yaml.safe_load(f)

    return cluster_config


def get_requirements(requirements_path: str) -> str:
    with open(requirements_path) as f:
        return f.read()


def _resource(name: str, region: str) -> Any:
    import boto3
    from botocore.config import Config

    boto_config = Config(retries={"max_attempts": BOTO_MAX_RETRIES})
    return boto3.resource(name, region, config=boto_config)


def _client(name: str, region: str) -> Any:
    return _resource(name, region).meta.client


def _get_role(role_name: str, region: str) -> Any:
    import botocore

    iam = _resource("iam", region)
    role = iam.Role(role_name)
    try:
        role.load()
        return role
    except botocore.exceptions.ClientError as exc:
        if exc.response.get("Error", {}).get("Code") == "NoSuchEntity":
            return None
        else:
            raise exc


def _get_user(user_name: str, region: str) -> Any:
    import botocore

    iam = _resource("iam", region)
    user = iam.User(user_name)
    try:
        user.load()
        return user
    except botocore.exceptions.ClientError as exc:
        if exc.response.get("Error", {}).get("Code") == "NoSuchEntity":
            return None
        else:
            raise exc


def get_available_regions() -> List[str]:
    import boto3

    client = boto3.client("ec2")
    return [region["RegionName"] for region in client.describe_regions()["Regions"]]


def launch_gcp_cloud_setup(name: str, region: str) -> None:
    # TODO: enforce uniqueness for user's clouds
    quote_safe_name = quote_sanitize(name, safe="")
    query_params = urlencode({"region": region})
    # TODO: Replace this with a proper endpoint
    endpoint = f"/api/v2/clouds/gcp/create/{quote_safe_name}?{query_params}"
    webbrowser.open(get_endpoint(endpoint))


class Timer:
    """
    Code adopted from https://stackoverflow.com/a/39504463/3727678
    Spawn thread and time process that may be blocking.
    """

    def timer_generator(self) -> Iterator[str]:
        while True:
            time_diff = time.gmtime(time.time() - self.start_time)
            yield "{0}: {1}".format(self.message, time.strftime("%M:%S", time_diff))

    def __init__(self, message: str = "") -> None:
        self.message = message
        self.busy = False
        self.start_time = 0.0

    def timer_task(self) -> None:
        while self.busy:
            sys.stdout.write(next(self.timer_generator()))
            sys.stdout.flush()
            time.sleep(0.1)
            sys.stdout.write("\b" * (len(self.message) + 20))
            sys.stdout.flush()

    def start(self) -> None:
        self.busy = True
        self.start_time = time.time()
        threading.Thread(target=self.timer_task).start()

    def stop(self) -> None:
        sys.stdout.write("\n")
        sys.stdout.flush()
        self.busy = False
        self.start_time = 0.0
        time.sleep(1)


def check_is_feature_flag_on(flag_key: str, default: bool = False) -> bool:
    try:
        use_snapshot_resp = send_json_request(
            "/api/v2/userinfo/check_is_feature_flag_on", {"flag_key": flag_key},
        )
    except Exception:
        return default

    return cast(bool, use_snapshot_resp["result"]["is_on"])


def get_active_sessions(
    project_id: str, session_name: str, api_client: Optional[DefaultApi]
) -> Any:
    if api_client:
        return api_client.list_sessions_api_v2_sessions_get(
            project_id=project_id, name=session_name, active_only=True
        ).results
    else:
        response = anyscale.util.send_json_request(
            "/api/v2/sessions/",
            {"project_id": project_id, "name": session_name, "active_only": True},
        )
        return response["results"]


def get_project_directory_name(project_id: str, api_client: DefaultApi = None) -> str:
    if api_client is None:
        api_client = get_api_client()

    # TODO (yiran): return error early if project doesn't exist.
    with format_api_exception(ApiException):
        resp = api_client.get_project_api_v2_projects_project_id_get(project_id)
    directory_name = resp.result.directory_name
    assert len(directory_name) > 0, "Empty directory name found."
    return cast(str, directory_name)


def get_working_dir(
    cluster_config: Dict[str, Any], project_id: str, api_client: DefaultApi = None
) -> str:
    working_dir: Optional[str] = (
        cluster_config.get("metadata", {}).get("anyscale", {}).get("working_dir")
    )
    if working_dir:
        return working_dir
    else:
        return f"/home/ray/{get_project_directory_name(project_id, api_client)}"


def get_wheel_url(
    ray_commit: str,
    ray_version: str,
    py_version: Optional[str] = None,
    sys_platform: Optional[str] = None,
) -> str:
    """Return S3 URL for the given release spec or 'latest'."""
    if py_version is None:
        py_version = "".join(str(x) for x in sys.version_info[0:2])
    if sys_platform is None:
        sys_platform = sys.platform

    if sys_platform == "darwin":
        if py_version == "38":
            platform = "macosx_10_13_x86_64"
        else:
            platform = "macosx_10_13_intel"
    elif sys_platform == "win32":
        platform = "win_amd64"
    else:
        platform = "manylinux2014_x86_64"

    if py_version == "38":
        py_version_malloc = py_version
    else:
        py_version_malloc = f"{py_version}m"

    if "dev" in ray_version:
        ray_release = f"master/{ray_commit}"
    else:
        ray_release = f"releases/{ray_version}/{ray_commit}"
    return (
        "https://s3-us-west-2.amazonaws.com/ray-wheels/"
        "{}/ray-{}-cp{}-cp{}-{}.whl".format(
            ray_release, ray_version, py_version, py_version_malloc, platform
        )
    )


def wait_for_session_start(
    project_id: str, session_name: str, api_client: Optional[DefaultApi] = None
) -> str:
    print(f"Waiting for cluster {session_name} to start.")
    if api_client is None:
        api_client = get_api_client()

    while True:
        sessions = api_client.list_sessions_api_v2_sessions_get(
            project_id=project_id, name=session_name, active_only=False
        ).results

        if len(sessions) > 0:
            session = sessions[0]

            if session.state == "Running" and session.pending_state is None:
                return cast(str, session.id)

            # Check for start up errors
            if (
                session.state_data
                and session.state_data.startup
                and session.state_data.startup.startup_error
            ):
                raise click.ClickException(
                    f"Error while starting session {session_name}: {session.state_data.startup.startup_error}"
                )
            elif (
                session.state
                and "Errored" in session.state
                and session.pending_state is None
            ):
                raise click.ClickException(
                    f"Error while starting session {session_name}: Session startup failed due to an error ({session.state})."
                )
            elif (
                session.state
                and session.state in {"Terminated", "Stopped"}
                and session.pending_state is None
            ):
                # Session is created in Terminated state; Check pending state to see if it is pending transition.
                raise click.ClickException(
                    f"Error while starting session {session_name}: Session is still in stopped/terminated state."
                )
        else:
            raise click.ClickException(
                f"Error while starting session {session_name}: Session doesn't exist."
            )

        time.sleep(5)


def wait_for_head_node_start(
    project_id: str,
    session_name: str,
    session_id: str,
    api_client: Optional[DefaultApi] = None,
) -> str:
    url = get_endpoint(f"/projects/{project_id}/sessions/{session_id}")
    print(
        f"Waiting for head node of session {session_name} to start. "
        f"Please view the startup logs at {url}"
    )
    if api_client is None:
        api_client = get_api_client()

    retry_count = 0
    while True:
        sessions = api_client.list_sessions_api_v2_sessions_get(
            project_id=project_id, name=session_name, active_only=False
        ).results

        if len(sessions) > 0:
            session = sessions[0]

            if (
                session.state == "Running" and session.pending_state is None
            ) or session.state == "AwaitingFileMounts":
                return cast(str, session.id)

            # Check for start up errors
            if (
                session.state_data
                and session.state_data.startup
                and session.state_data.startup.startup_error
            ):
                raise click.ClickException(
                    f"Error while starting session {session_name}: {session.state_data.startup.startup_error}"
                )
            elif session.state and "Errored" in session.state:
                raise click.ClickException(
                    f"Error while starting session {session_name}: Session startup failed due to an error ({session.state})."
                )
            elif (
                retry_count >= 2
                and session.state
                and session.state in {"Terminated", "Stopped"}
                and session.pending_state is None
            ):
                # Session is created in Terminated state; Ignore the first few state checks.
                raise click.ClickException(
                    f"Error while starting session {session_name}: Session is still in stopped/terminated state."
                )
        else:
            raise click.ClickException(
                f"Error while starting session {session_name}: Session doesn't exist."
            )

        time.sleep(5)
        retry_count += 1


def download_anyscale_wheel(api_client: DefaultApi, session_id: str) -> None:
    with format_api_exception(ApiException):
        wheel_resp = api_client.session_get_anyscale_wheel_api_v2_sessions_session_id_anyscale_wheel_get(
            session_id=session_id, _preload_content=False
        )
    wheel_path_raw = wheel_resp.headers["content-disposition"]
    assert "filename" in wheel_path_raw, "Error getting anyscale wheel"
    wheel_path = wheel_path_raw.split("filename=")[1].strip('"')
    os.makedirs(os.path.dirname(wheel_path), exist_ok=True)
    with open(wheel_path, "wb+") as f:
        f.write(wheel_resp.data)
        f.flush()


def validate_cluster_configuration(
    cluster_config_file_name: str,
    cluster_config: Optional[DefaultApi] = None,
    api_instance: Optional[DefaultApi] = None,
) -> None:
    assert api_instance

    if not os.path.isfile(cluster_config_file_name):
        raise click.ClickException(
            "The configuration file {} does not exist. Please provide a valid config file.".format(
                cluster_config_file_name
            )
        )

    if not cluster_config:
        try:
            with open(cluster_config_file_name) as f:
                cluster_config = yaml.safe_load(f)
        except (ValueError, yaml.YAMLError):
            raise click.ClickException(
                "\tThe configuration file {} does not have a valid format. "
                "\n\tPlease look at https://github.com/ray-project/ray/blob/master/python/ray/autoscaler/aws/example-full.yaml "
                "for an example configuration file.".format(cluster_config_file_name)
            )

    try:
        api_instance.validate_cluster_api_v2_sessions_validate_cluster_post(
            body={"config": json.dumps(cluster_config)}
        )
    except ApiException as e:
        error = json.loads(json.loads(e.body)["error"]["detail"])
        path = ".".join(error["path"])
        if error["path"]:
            formatted_error = 'Error occured at "{k}: {v}" because {message}.\nSchema description for {k}:\n{schema}'.format(
                k=path,
                v=error["instance"],
                message=error["message"],
                schema=json.dumps(error["schema"], indent=4, sort_keys=True),
            )
        else:
            formatted_error = 'Error occured at "{v}" because {message}.\nSchema description:\n{schema}'.format(
                v=error["instance"],
                message=error["message"],
                schema=json.dumps(error["schema"], indent=4, sort_keys=True),
            )
        raise click.ClickException(
            "The configuration file {0} is not valid.\n{1}".format(
                cluster_config_file_name, formatted_error
            )
        )


def _get_rsync_args(rsync_exclude: List[str], rsync_filter: List[str]) -> List[str]:
    rsync_exclude_args = [["--exclude", exclude] for exclude in rsync_exclude]
    rsync_filter_args = [["--filter", f"dir-merge,- {f}"] for f in rsync_filter]

    # Combine and flatten the two lists
    return [
        arg for sublist in rsync_exclude_args + rsync_filter_args for arg in sublist
    ]


def get_rsync_command(
    ssh_command: List[str],
    source: str,
    ssh_user: str,
    head_ip: str,
    target: str,
    delete: bool,
    rsync_exclude: Optional[List[str]] = None,
    rsync_filter: Optional[List[str]] = None,
    dry_run: bool = False,
) -> Tuple[List[str], Optional[Dict[str, Any]]]:
    rsync_executable = "rsync"
    env = None
    if rsync_exclude is None:
        rsync_exclude = []
    if rsync_filter is None:
        rsync_filter = []

    command_list = [
        rsync_executable,
        "--rsh",
        " ".join(ssh_command),
        "-avz",
    ]
    command_list += _get_rsync_args(rsync_exclude, rsync_filter)

    if delete:
        # Deletes files in target that doesn't exist in source
        command_list.append("--delete")

    if dry_run:
        command_list += [
            "--dry-run",
            "--itemize-changes",
            "--out-format",
            "%o %f",
        ]

    command_list += [
        source,
        "{}@{}:{}".format(ssh_user, head_ip, target),
    ]
    return command_list, env


def populate_session_args(cluster_config_str: str, config_file_name: str) -> str:
    import jinja2

    env = jinja2.Environment()
    t = env.parse(cluster_config_str)
    for elem in t.body[0].nodes:

        if isinstance(elem, jinja2.nodes.Getattr):  # type: ignore
            if elem.attr not in os.environ:
                prefixed_command = " ".join(
                    [f"{elem.attr}=<value>", "anyscale"] + sys.argv[1:]
                )
                raise click.ClickException(
                    f"\tThe environment variable {elem.attr} was not set, yet it is required "
                    f"for configuration file {config_file_name}.\n\tPlease specify {elem.attr} "
                    f"by prefixing the command.\n\t\t{prefixed_command}"
                )

    template = jinja2.Template(cluster_config_str)
    cluster_config_filled = template.render(env=os.environ)
    return cluster_config_filled


def canonicalize_remote_location(
    cluster_config: Dict[str, Any], remote_location: Optional[str], project_id: str
) -> Optional[str]:
    """Returns remote_location, but changes it from being based
    in "~/", "/root/" or "/ray" to match working_dir
    """
    # Include the /root path to ensure that absolute paths also work
    # This is because of an implementation detail in OSS Ray's rsync
    if bool(get_container_name(cluster_config)) and bool(remote_location):
        remote_location = str(remote_location)
        working_dir = get_working_dir(cluster_config, project_id)
        for possible_name in ["root", "ray"]:
            full_name = f"/{possible_name}/"
            # TODO(ilr) upstream this to OSS Ray
            if working_dir.startswith("~/") and remote_location.startswith(full_name):
                return remote_location.replace(full_name, "~/", 1)

            if working_dir.startswith(full_name) and remote_location.startswith("~/"):
                return remote_location.replace("~/", full_name, 1)

    return remote_location


@contextmanager
def format_api_exception(exception_type: Type[Exception]) -> Iterator[None]:
    # This should always be called with exceptions that have a status and reason attribute
    try:
        yield
    except exception_type as e:
        if isinstance(e, ApiException):
            raise click.ClickException(f"{e.status}: {str(e)}")

        raise click.ClickException(
            "{0}: {1}".format(e.status, json.loads(e.body)["error"]["detail"])  # type: ignore
        )


def init_sentry(ctx: click.core.Context) -> None:
    # Only report production and staging errors
    if shared_anyscale_conf.ANYSCALE_ENV not in ["production", "staging"]:
        return

    import sentry_sdk

    user_info = get_user_info()
    if user_info:
        if user_info.email in [
            "test@anyscale.com",
            "ci-staging@anyscale.com",
            "ci-production@anyscale.com",
        ]:
            return
        sentry_sdk.set_user(
            {
                "id": user_info.id,
                "email": user_info.email,
                "username": user_info.username,
            }
        )

    sentry_sdk.set_tag("command", ctx.invoked_subcommand)
    sentry_sdk.init(
        "https://68d7689b55024e91bb31694539111baa@o565360.ingest.sentry.io/5706818",
        release=get_anyscale_version(),
        environment=shared_anyscale_conf.ANYSCALE_ENV,
    )


def get_user_info() -> Optional[UserInfo]:
    try:
        api_client = get_api_client()
    except click.exceptions.ClickException:
        return None
    return api_client.get_user_info_api_v2_userinfo_get().result


def get_anyscale_version() -> str:
    # Return git sha if anyscale pip package was built in development mode
    from pip._internal.operations.freeze import FrozenRequirement
    import pkg_resources

    distributions = {v.key: v for v in pkg_resources.working_set}
    distribution = distributions["anyscale"]
    frozen_requirement = FrozenRequirement.from_dist(distribution)  # type: ignore
    if frozen_requirement.editable:
        try:
            return (
                subprocess.check_output(  # noqa: B1
                    ["git", "describe", "--always"],
                    cwd=os.path.dirname(os.path.realpath(__file__)),
                )  # noqa: B1
                .strip()
                .decode("utf-8")
            )
        except subprocess.CalledProcessError:
            print("Not a git repository, will use standard release versioning")
    return anyscale.__version__



from typing import Any

import anyscale.client.openapi_client as openapi_client
from anyscale.client.openapi_client.api.default_api import DefaultApi
import anyscale.conf
from anyscale.credentials import load_credentials
import anyscale.sdk.anyscale_client as anyscale_client
from anyscale.sdk.anyscale_client.api.default_api import DefaultApi as AnyscaleApi
import anyscale.shared_anyscale_utils.conf as shared_anyscale_conf
from anyscale.version import __version__ as version


def instantiate_api_client(
    no_cli_token: bool = False, cli_token: str = "", host: str = ""
) -> DefaultApi:
    """
    Instantiates client to interact with our frontend APIs
    """
    if (cli_token and not host) or (not cli_token and host):
        raise ValueError("Both cli_token and host need to be provided.")

    if not no_cli_token and anyscale.conf.CLI_TOKEN is None and not cli_token:
        anyscale.conf.CLI_TOKEN = load_credentials()
    configuration = openapi_client.Configuration(
        host=host or shared_anyscale_conf.ANYSCALE_HOST
    )
    configuration.connection_pool_maxsize = 100

    if no_cli_token:
        api_client = openapi_client.ApiClient(configuration)
    else:
        cookie = (
            f"cli_token={cli_token}"
            if cli_token
            else f"cli_token={anyscale.conf.CLI_TOKEN}"
        )
        api_client = openapi_client.ApiClient(configuration, cookie=cookie)
    configure_open_api_client_headers(api_client, "CLI")
    api_instance = openapi_client.DefaultApi(api_client)
    return api_instance


def instantiate_anyscale_client() -> AnyscaleApi:
    """
    Instantiates client to interact with our externalized APIs
    """

    if anyscale.conf.CLI_TOKEN is None:
        anyscale.conf.CLI_TOKEN = load_credentials()

    configuration = anyscale_client.Configuration(
        host=shared_anyscale_conf.ANYSCALE_HOST + "/ext"
    )
    configuration.connection_pool_maxsize = 100

    api_client = anyscale_client.ApiClient(
        configuration, cookie=f"cli_token={anyscale.conf.CLI_TOKEN}"
    )
    configure_open_api_client_headers(api_client, "CLI")

    api_instance = anyscale_client.DefaultApi(api_client)
    return api_instance


def get_api_client() -> DefaultApi:
    if _api_client.api_client is None:
        _api_client.api_client = instantiate_api_client()

    return _api_client.api_client


def get_anyscale_api_client() -> AnyscaleApi:
    if _api_client.anyscale_client is None:
        _api_client.anyscale_client = instantiate_anyscale_client()

    return _api_client.anyscale_client


# client is of type APIClient, which is auto-generated
def configure_open_api_client_headers(client: Any, client_name: str) -> None:
    client.set_default_header("X-Anyscale-Client", client_name)
    client.set_default_header("X-Anyscale-Client-Version", version)


class _ApiClient(object):
    api_client: DefaultApi = None
    anyscale_client: AnyscaleApi = None


_api_client = _ApiClient()



# coding: utf-8

"""
    Anyscale API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from anyscale_client.configuration import Configuration


class Actor(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'id': 'str',
        'ray_actor_id': 'str',
        'name': 'str',
        'status': 'ActorStatus',
        'lifetime': 'ActorLifetime',
        'class_name': 'str',
        'task_name': 'str',
        'job_id': 'str',
        'created_at': 'datetime',
        'finished_at': 'datetime',
        'current_worker_id': 'str',
        'runtime_environment_id': 'str'
    }

    attribute_map = {
        'id': 'id',
        'ray_actor_id': 'ray_actor_id',
        'name': 'name',
        'status': 'status',
        'lifetime': 'lifetime',
        'class_name': 'class_name',
        'task_name': 'task_name',
        'job_id': 'job_id',
        'created_at': 'created_at',
        'finished_at': 'finished_at',
        'current_worker_id': 'current_worker_id',
        'runtime_environment_id': 'runtime_environment_id'
    }

    def __init__(self, id=None, ray_actor_id=None, name=None, status=None, lifetime=None, class_name=None, task_name=None, job_id=None, created_at=None, finished_at=None, current_worker_id=None, runtime_environment_id=None, local_vars_configuration=None):  # noqa: E501
        """Actor - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._id = None
        self._ray_actor_id = None
        self._name = None
        self._status = None
        self._lifetime = None
        self._class_name = None
        self._task_name = None
        self._job_id = None
        self._created_at = None
        self._finished_at = None
        self._current_worker_id = None
        self._runtime_environment_id = None
        self.discriminator = None

        self.id = id
        self.ray_actor_id = ray_actor_id
        if name is not None:
            self.name = name
        self.status = status
        self.lifetime = lifetime
        if class_name is not None:
            self.class_name = class_name
        if task_name is not None:
            self.task_name = task_name
        self.job_id = job_id
        self.created_at = created_at
        if finished_at is not None:
            self.finished_at = finished_at
        self.current_worker_id = current_worker_id
        self.runtime_environment_id = runtime_environment_id

    @property
    def id(self):
        """Gets the id of this Actor.  # noqa: E501


        :return: The id of this Actor.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this Actor.


        :param id: The id of this Actor.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def ray_actor_id(self):
        """Gets the ray_actor_id of this Actor.  # noqa: E501


        :return: The ray_actor_id of this Actor.  # noqa: E501
        :rtype: str
        """
        return self._ray_actor_id

    @ray_actor_id.setter
    def ray_actor_id(self, ray_actor_id):
        """Sets the ray_actor_id of this Actor.


        :param ray_actor_id: The ray_actor_id of this Actor.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and ray_actor_id is None:  # noqa: E501
            raise ValueError("Invalid value for `ray_actor_id`, must not be `None`")  # noqa: E501

        self._ray_actor_id = ray_actor_id

    @property
    def name(self):
        """Gets the name of this Actor.  # noqa: E501


        :return: The name of this Actor.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this Actor.


        :param name: The name of this Actor.  # noqa: E501
        :type: str
        """

        self._name = name

    @property
    def status(self):
        """Gets the status of this Actor.  # noqa: E501


        :return: The status of this Actor.  # noqa: E501
        :rtype: ActorStatus
        """
        return self._status

    @status.setter
    def status(self, status):
        """Sets the status of this Actor.


        :param status: The status of this Actor.  # noqa: E501
        :type: ActorStatus
        """
        if self.local_vars_configuration.client_side_validation and status is None:  # noqa: E501
            raise ValueError("Invalid value for `status`, must not be `None`")  # noqa: E501

        self._status = status

    @property
    def lifetime(self):
        """Gets the lifetime of this Actor.  # noqa: E501


        :return: The lifetime of this Actor.  # noqa: E501
        :rtype: ActorLifetime
        """
        return self._lifetime

    @lifetime.setter
    def lifetime(self, lifetime):
        """Sets the lifetime of this Actor.


        :param lifetime: The lifetime of this Actor.  # noqa: E501
        :type: ActorLifetime
        """
        if self.local_vars_configuration.client_side_validation and lifetime is None:  # noqa: E501
            raise ValueError("Invalid value for `lifetime`, must not be `None`")  # noqa: E501

        self._lifetime = lifetime

    @property
    def class_name(self):
        """Gets the class_name of this Actor.  # noqa: E501


        :return: The class_name of this Actor.  # noqa: E501
        :rtype: str
        """
        return self._class_name

    @class_name.setter
    def class_name(self, class_name):
        """Sets the class_name of this Actor.


        :param class_name: The class_name of this Actor.  # noqa: E501
        :type: str
        """

        self._class_name = class_name

    @property
    def task_name(self):
        """Gets the task_name of this Actor.  # noqa: E501


        :return: The task_name of this Actor.  # noqa: E501
        :rtype: str
        """
        return self._task_name

    @task_name.setter
    def task_name(self, task_name):
        """Sets the task_name of this Actor.


        :param task_name: The task_name of this Actor.  # noqa: E501
        :type: str
        """

        self._task_name = task_name

    @property
    def job_id(self):
        """Gets the job_id of this Actor.  # noqa: E501


        :return: The job_id of this Actor.  # noqa: E501
        :rtype: str
        """
        return self._job_id

    @job_id.setter
    def job_id(self, job_id):
        """Sets the job_id of this Actor.


        :param job_id: The job_id of this Actor.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and job_id is None:  # noqa: E501
            raise ValueError("Invalid value for `job_id`, must not be `None`")  # noqa: E501

        self._job_id = job_id

    @property
    def created_at(self):
        """Gets the created_at of this Actor.  # noqa: E501


        :return: The created_at of this Actor.  # noqa: E501
        :rtype: datetime
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this Actor.


        :param created_at: The created_at of this Actor.  # noqa: E501
        :type: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_at is None:  # noqa: E501
            raise ValueError("Invalid value for `created_at`, must not be `None`")  # noqa: E501

        self._created_at = created_at

    @property
    def finished_at(self):
        """Gets the finished_at of this Actor.  # noqa: E501


        :return: The finished_at of this Actor.  # noqa: E501
        :rtype: datetime
        """
        return self._finished_at

    @finished_at.setter
    def finished_at(self, finished_at):
        """Sets the finished_at of this Actor.


        :param finished_at: The finished_at of this Actor.  # noqa: E501
        :type: datetime
        """

        self._finished_at = finished_at

    @property
    def current_worker_id(self):
        """Gets the current_worker_id of this Actor.  # noqa: E501


        :return: The current_worker_id of this Actor.  # noqa: E501
        :rtype: str
        """
        return self._current_worker_id

    @current_worker_id.setter
    def current_worker_id(self, current_worker_id):
        """Sets the current_worker_id of this Actor.


        :param current_worker_id: The current_worker_id of this Actor.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and current_worker_id is None:  # noqa: E501
            raise ValueError("Invalid value for `current_worker_id`, must not be `None`")  # noqa: E501

        self._current_worker_id = current_worker_id

    @property
    def runtime_environment_id(self):
        """Gets the runtime_environment_id of this Actor.  # noqa: E501


        :return: The runtime_environment_id of this Actor.  # noqa: E501
        :rtype: str
        """
        return self._runtime_environment_id

    @runtime_environment_id.setter
    def runtime_environment_id(self, runtime_environment_id):
        """Sets the runtime_environment_id of this Actor.


        :param runtime_environment_id: The runtime_environment_id of this Actor.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and runtime_environment_id is None:  # noqa: E501
            raise ValueError("Invalid value for `runtime_environment_id`, must not be `None`")  # noqa: E501

        self._runtime_environment_id = runtime_environment_id

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, Actor):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, Actor):
            return True

        return self.to_dict() != other.to_dict()



from typing import List, Tuple

import click

from anyscale.controllers.exec_controller import ExecController


@click.command(name="exec", help="Execute shell commands in interactive session.")
@click.option(
    "--session-name",
    "-n",
    type=str,
    required=False,
    default=None,
    help="Session name optional if only one running session.",
    envvar="SESSION_NAME",
)
@click.option(
    "--screen", is_flag=True, default=False, help="Run the command in a screen."
)
@click.option("--tmux", is_flag=True, default=False, help="Run the command in tmux.")
@click.option(
    "--port-forward",
    "-p",
    required=False,
    multiple=True,
    type=int,
    help="Port to forward. Use this multiple times to forward multiple ports.",
)
@click.option(
    "--sync",
    is_flag=True,
    default=False,
    help="Rsync all the file mounts before executing the command.",
)
@click.option(
    "--stop",
    help="Stop session after command finishes executing.",
    is_flag=True,
    default=False,
)
@click.option(
    "--terminate",
    help="Terminate session after command finishes executing.",
    is_flag=True,
    default=False,
)
@click.argument("commands", nargs=-1, type=str)
def anyscale_exec(
    session_name: str,
    screen: bool,
    tmux: bool,
    port_forward: Tuple[int],
    sync: bool,
    stop: bool,
    terminate: bool,
    commands: List[str],
) -> None:
    exec_controller = ExecController()
    exec_controller.anyscale_exec(
        session_name, screen, tmux, port_forward, sync, stop, terminate, commands
    )
